/*
	bzlibmodule.c  :  A simple Python interface to Julian Sewards bzip2
	                           compression system

	0.02b alpha (25 April 1999)  Copyright Laurence Tratt 1999



  Licence:

	Copyright Laurence Tratt 1999

	All rights reserved

	Permission to use, copy, modify, and distribute this software and its
	documentation for any purpose and without fee is hereby granted, provided that
	the above copyright notice appear in all copies and that both that copyright
	notice and this permission notice appear in supporting documentation.

	THE AUTHOR - LAURENCE TRATT - DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
	SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
	NO EVENT SHALL THE AUTHOR FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
	ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
	AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
	IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.



  Version history:

	0.01 alpha (19 April 1999)
	0.02 alpha (20 April 1999)
	    Corrected C function names (they started PyZlib instead of PyBZlib. Oops)
	0.02b alpha (25 April 1999)
        Tidied up installation procedure and instructions. There is no change to
          any code



  Contact:

	Laurence Tratt    tratt@dcs.kcl.ac.uk



  See:

 	bzip2 source code and documentation can be obtained from its homepage at
 	http://www.muraroa.demon.co.uk/
 */






#define DEFAULT_DECOMPRESS_SIZE 100 * 1024
#define DEFAULT_DECOMPRESS_INPUT_MULTIPLE 4


#include "Python.h"
#include "math.h"

#include "bzlib.h"





static PyObject *BZlibError;




static char compress__doc__[] = "compress(string, block size)";


static PyObject * PyBZlib_compress(self, args)
	PyObject *self;
	PyObject *args;
{
	PyObject *return_value;
	char *input, *output;
	int err, block_size;
	unsigned int input__length, output__length;

	if (! PyArg_ParseTuple(args, "s#i", &input, &input__length, &block_size))
		return NULL;

	output__length = input__length + ceil(input__length * (1 / 100.0)) + 600;

	output = malloc(output__length);
	if (output == NULL)
	{
		PyErr_SetString(PyExc_MemoryError, "Can't allocate memory to compress data");
		return NULL;
	}

	err = bzBuffToBuffCompress(output, &output__length, input, input__length, block_size, 0, 0);
	switch (err)
	{
		case (BZ_MEM_ERROR):
			PyErr_SetString(PyExc_MemoryError, "Insufficient memory to compress data");
			free(output);
			return NULL;

		case (BZ_PARAM_ERROR):
			PyErr_SetString(BZlibError, "Invalid blocksize parameter");
			free(output);
			return NULL;
	}

	return_value = PyString_FromStringAndSize(output, output__length);
	free(output);

	return return_value;
}





static char decompress__doc__[] = "decompress(string, [small])";

static PyObject * PyBZlib_decompress(self, args)
	PyObject *self;
	PyObject *args;
{
	PyObject *return_value;
	char *input;
	int err, small, buffer_size, input__length;
	bz_stream strm;

	small = 0;

	if (! PyArg_ParseTuple(args, "s#|i", &input, &input__length, &small))
		return NULL;

	strm.bzalloc = NULL;
	strm.bzfree = NULL;
	strm.opaque = NULL;

	err = bzDecompressInit(&strm, 0, small);
	switch (err)
	{
		case (BZ_MEM_ERROR):
			PyErr_SetString(PyExc_MemoryError, "Insufficient memory to decompress data");
			return NULL;

		case (BZ_PARAM_ERROR):
			PyErr_SetString(BZlibError, "Invalid smaller parameter");
			return NULL;
	}

	strm.next_in = input;
	strm.avail_in = input__length;
	strm.total_in = 0;
	if (input__length * DEFAULT_DECOMPRESS_INPUT_MULTIPLE < DEFAULT_DECOMPRESS_SIZE)
		buffer_size = input__length * DEFAULT_DECOMPRESS_INPUT_MULTIPLE;
	else
		buffer_size = DEFAULT_DECOMPRESS_SIZE;
	strm.next_out = malloc(buffer_size);
	strm.avail_out = buffer_size;
	if (strm.next_out == NULL)
	{
		PyErr_SetString(PyExc_MemoryError, "Can't allocate enough memory to compress data");
		bzDecompressEnd(&strm);
		return NULL;
	}

	do
	{
		err = bzDecompress(&strm);
		switch (err)
		{
			case (BZ_DATA_ERROR):
				PyErr_SetString(BZlibError, "Data integrity error when decompressing");
				bzDecompressEnd(&strm);
				free(strm.next_out);
				return NULL;

			case (BZ_DATA_ERROR_MAGIC):
				PyErr_SetString(BZlibError, "Data does not begin with the correct magic bytes");
				bzDecompressEnd(&strm);
				free(strm.next_out);
				return NULL;

			case (BZ_MEM_ERROR):
				PyErr_SetString(BZlibError, "Can't allocate enough memory to decompress data");
				bzDecompressEnd(&strm);
				free(strm.next_out);
				return NULL;
		}

		if (err != BZ_STREAM_END)
		{
			if ((input__length - strm.total_in) * DEFAULT_DECOMPRESS_INPUT_MULTIPLE < DEFAULT_DECOMPRESS_SIZE)
				buffer_size = (input__length - strm.total_in) * DEFAULT_DECOMPRESS_INPUT_MULTIPLE;
			else
				buffer_size = DEFAULT_DECOMPRESS_SIZE;
			strm.next_out = realloc(strm.next_out - strm.total_out, strm.total_out + buffer_size) + strm.total_out;
			strm.avail_out = buffer_size;
			if (strm.next_out == NULL)
			{
				PyErr_SetString(PyExc_MemoryError, "Can't allocate enough memory to compress data");
				bzDecompressEnd(&strm);
				return NULL;
			}
		}
	} while (err != BZ_STREAM_END);
	bzDecompressEnd(&strm);

	return_value = PyString_FromStringAndSize(strm.next_out - strm.total_out, strm.total_out);
	free(strm.next_out - strm.total_out);

	return return_value;
}




static PyMethodDef bzlib_methods[] =
{
        {"compress", (PyCFunction) PyBZlib_compress, 1, compress__doc__},
        {"decompress", (PyCFunction) PyBZlib_decompress, 1, decompress__doc__},
        {NULL, NULL}
};





static char bzlib_module_documentation[] = "The functions in this module allow compression and decompression using the bzlib library";

DL_EXPORT(void)
initbzlib()
{
	PyObject *module, *dict;

	module = Py_InitModule4("bzlib", bzlib_methods, bzlib_module_documentation, (PyObject*) NULL, PYTHON_API_VERSION);
	dict = PyModule_GetDict(module);

	BZlibError = PyErr_NewException("bzlib.error", NULL, NULL);
	PyDict_SetItemString(dict, "error", BZlibError);
}